/* wptPassphraseCB.cpp - GPGME Passphrase Callback
 *	Copyright (C) 2001, 2002, 2003, 2005, 2006 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <ctype.h>

#include "resource.h"
#include "wptNLS.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptContext.h"
#include "wptDlgs.h"
#include "wptErrors.h"
#include "wptTypes.h"
#include "wptKeylist.h"
#include "wptAgent.h"
#include "wptRegistry.h"
#include "wptUTF8.h"

const char* get_symkey_algo (int algo);

#define item_ctrl_id(cmd) \
    ((cmd) == GPG_CMD_DECRYPT? IDC_DECRYPT_PWD : IDC_DECRYPT_SIGN_PWD)

#define item_ctrl_id2(cmd) \
    ((cmd) == GPG_CMD_DECRYPT? IDC_DECRYPT_HIDE : IDC_DECRYPT_SIGN_HIDE)


void ListBox_AddString_utf8 (HWND lb, const char *txt);

/* Overwrite passphrase and free memory. */
static void
burn_passphrase (char **pwd)
{
    char *pass = *pwd;
    wipememory (pass, strlen (pass));
    delete []pass;
    *pwd = NULL;
}


/* Dialog procedure for the passphrase callback. */
static BOOL CALLBACK
passphrase_callback_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{    
    static passphrase_cb_s *c;
    gpgme_decrypt_result_t res=NULL;
    gpgme_sign_result_t res_sig=NULL;
    gpgme_recipient_t recip=NULL, r;
    winpt_key_s key;
    void *item;
    const char *id;
    char *info;
    int n;

    switch (msg) {
    case WM_ACTIVATE:
	safe_edit_control_init (dlg, item_ctrl_id (c->gpg_cmd));
	break;

    case WM_DESTROY:
	safe_edit_control_free (dlg, item_ctrl_id (c->gpg_cmd));
	break;

    case WM_INITDIALOG:
	c = (passphrase_cb_s *)lparam;
	if (!c)
	    BUG (0);
	SetDlgItemText (dlg, IDCANCEL, _("&Cancel"));
	SetWindowText (dlg, c->title);
	if (c->gpg_cmd == GPG_CMD_DECRYPT) {
	    SetDlgItemText (dlg, IDC_DECRYPT_HIDE, _("&Hide Typing"));
	    SetDlgItemText (dlg, IDC_DECRYPT_LISTINF,
			    _("Encrypted with the following public key(s)"));
	    CheckDlgButton (dlg, IDC_DECRYPT_HIDE, BST_CHECKED);
	}
	else if (c->gpg_cmd == GPG_CMD_SIGN) {
	    SetDlgItemText (dlg, IDC_DECRYPT_SIGN_HIDE, _("&Hide Typing"));
	    CheckDlgButton (dlg, IDC_DECRYPT_SIGN_HIDE, BST_CHECKED);
	}
	/* Because it depends on the order the keys are stored in the
	   keyring whether res->recipients is complete or not, we also
	   support that the recipients were externally extracted and then
	   we use this list. */
	if (c->recipients)
	    recip = c->recipients; /* recipients were already extracted. */
	else {
	    res = gpgme_op_decrypt_result (c->gpg);
	    if (res && res->recipients)
		recip = res->recipients;
	}
	if (recip != NULL && c->gpg_cmd == GPG_CMD_DECRYPT) {
	    for (r = recip; r; r = r->next) {
		memset (&key, 0, sizeof (key));
		if (!winpt_get_pubkey (r->keyid, &key)) {
		    gpgme_user_id_t u = key.ctx->uids;

		    id = u->name;
		    if (!id)
			id = _("Invalid User ID");
		    n = 32+strlen (id)+1+4+strlen (r->keyid)+1;
		    if (u->email)
			n += strlen (u->email)+1;
		    info = new char [n+1];
		    if (!info)
			BUG (NULL);
		    if (!u->email || strlen (u->email) < 1)
			sprintf (info, "%s (%s, 0x%s)", id,
				 get_key_pubalgo (r->pubkey_algo), r->keyid+8);
		    else
			sprintf (info, "%s <%s> (%s, 0x%s)", id, u->email,
				 get_key_pubalgo (r->pubkey_algo), r->keyid+8);
		}
		else {
		    info = new char [32 + strlen (r->keyid)+1 + 4];
		    if (!info)
			BUG (NULL);
		    sprintf (info, _("Unknown key ID (%s, 0x%s)"), 
			     get_key_pubalgo (r->pubkey_algo), r->keyid+8);
		}
		ListBox_AddString_utf8 (GetDlgItem (dlg, IDC_DECRYPT_LIST), info);
		free_if_alloc (info);
		winpt_release_pubkey (&key);
	    }
	}
	else if (c->gpg_cmd == GPG_CMD_DECRYPT)
	    EnableWindow (GetDlgItem (dlg, IDC_DECRYPT_LIST), FALSE);
	SetDlgItemText (dlg, c->gpg_cmd == GPG_CMD_DECRYPT?
			IDC_DECRYPT_PWDINFO : IDC_DECRYPT_SIGN_PWDINFO,
			c->bad_pwd? _("Bad passphrase; Enter passphrase again") :
			_("Please enter your passphrase"));
	if (c->gpg_cmd == GPG_CMD_DECRYPT) {
	    SetFocus (GetDlgItem (dlg, IDC_DECRYPT_PWD));
	    if (res && !res->recipients) {
		const char *s = _("Symmetric encryption.\n"
				  "%s encrypted data.");
		const char *alg = get_symkey_algo (c->sym.sym_algo);
		info = new char[strlen (s) + strlen (alg) + 2];
		if (!info)
		    BUG (NULL);
		sprintf (info, s, alg);
		SetDlgItemText (dlg, IDC_DECRYPT_MSG, info);
		free_if_alloc (info);
	    }
	    else
		SetDlgItemText (dlg, IDC_DECRYPT_MSG, c->info);
	}
	else {
	    SetFocus (GetDlgItem (dlg, IDC_DECRYPT_SIGN_PWD));
	    SetDlgItemText (dlg, IDC_DECRYPT_SIGN_MSG, c->info);
	}
	center_window (dlg, NULL);
	SetForegroundWindow (dlg);
	return FALSE;

	case WM_SYSCOMMAND:
	    if (LOWORD (wparam) == SC_CLOSE) {
		SetDlgItemText (dlg, item_ctrl_id (c->gpg_cmd), "");
		c->cancel = 1;
		EndDialog (dlg, TRUE);
	    }
	    break;

	case WM_COMMAND:
	    switch (HIWORD (wparam)) {
	    case BN_CLICKED:
		if  (LOWORD (wparam) == IDC_DECRYPT_HIDE
		    || LOWORD (wparam) == IDC_DECRYPT_SIGN_HIDE) {
		    HWND hwnd;
		    int hide = IsDlgButtonChecked (dlg, item_ctrl_id2 (c->gpg_cmd));
		    hwnd = GetDlgItem (dlg, item_ctrl_id (c->gpg_cmd));
		    SendMessage (hwnd, EM_SETPASSWORDCHAR, hide? '*' : 0, 0);
		    SetFocus (hwnd);
		}
	    }

	    switch (LOWORD (wparam)) {
	    case IDOK:	
		/* XXX: the item is even cached when the passphrase is not
		        correct, which means that the user needs to delete all
		        cached entries to continue. */
		if (c->pwd)
		    burn_passphrase (&c->pwd);
		n = item_get_text_length (dlg, item_ctrl_id (c->gpg_cmd));
		if (!n) {
		    c->pwd = new char[2];
		    if (!c->pwd)
			BUG (NULL);
		    strcpy (c->pwd, "");
		}
		else {
		    char *p = new char[n+2];
		    if (!p)
			BUG (NULL);
		    /* Get the passphrase and convert it utf8.
		       This does not just the us-ascii charset. */
		    SafeGetDlgItemText (dlg, item_ctrl_id (c->gpg_cmd), p, n+1);
		    c->pwd = native_to_utf8 (p);
		    sfree_if_alloc (p);
		}
		res = gpgme_op_decrypt_result (c->gpg);
		if (!res)
		    res_sig = gpgme_op_sign_result (c->gpg);		
		if (reg_prefs.cache_time > 0 && !c->is_card &&
		    (res || res_sig)) {
		    if (agent_get_cache (c->keyid, &item))			
			agent_unlock_cache_entry (&item);
		    else 
			agent_put_cache (c->keyid, c->pwd, reg_prefs.cache_time);
		}
		c->cancel = 0;
		EndDialog (dlg, TRUE);
		return TRUE;
		
	    case IDCANCEL:
		SetDlgItemText (dlg, item_ctrl_id (c->gpg_cmd), "");
		c->cancel = 1;
		EndDialog (dlg, FALSE);
		return FALSE;
	    }
	    break;
    }
    
    return FALSE;
}


/* Extract the main keyid from @pass_info.
   Return value: long main keyid or NULL for an error. */
static const char*
parse_gpg_keyid (const char *pass_info)
{
    static char keyid[16+1];
    
    /* XXX: check for leading alpha-chars? */
    if (strlen (pass_info) < 16) {
	log_debug ("parse_gpg_keyid: error '%s'\r\n", pass_info);
	return NULL;
    }
    /* the format of the desc buffer looks like this:
       request_keyid[16] main_keyid[16] keytype[1] keylength[4]
       we use the main keyid to use only one cache entry. */
    strncpy (keyid, pass_info+17, 16);
    keyid[16] = 0;
    return keyid;
}


/* Parse the information in @uid_hint and @pass_info to generate
   a input message for the user in @desc. */
static int
parse_gpg_description (const char *uid_hint, const char *pass_info,
		       char *desc, int size)
{
    gpgme_pubkey_algo_t algo;
    char usedkey[16+1];
    char mainkey[16+1];
    char *p, *uid;
    int n=0;

    algo = (gpgme_pubkey_algo_t)0;
    /* Each uid_hint contains a long key-ID so it is at least 16 bytes. */
    if (strlen (uid_hint) < 17) {
	*desc = 0;
	log_debug ("parse_gpg_description: error '%s'\r\n", uid_hint);
	return -1;
    }

    while (p = strsep ((char**)&pass_info, " ")) {
	switch (n++) {
	case 0: strncpy (mainkey, p, 16); mainkey[16] = 0; break;
	case 1: strncpy (usedkey, p, 16); usedkey[16] = 0; break;
	case 2: algo = (gpgme_pubkey_algo_t)atol (p); break;
	}
    }
    uid_hint += 16; /* skip keyid */
    uid_hint += 1;  /* space */

    uid = utf8_to_native (uid_hint);
    if (strcmp (usedkey, mainkey))
	_snprintf (desc, size-1, 
		   _("You need a passphrase to unlock the secret key for\n"
		     "user: \"%s\"\n"
		     "%s key, ID %s (main key ID %s)\n"),
		   uid, get_key_pubalgo (algo), usedkey+8, mainkey+8);
    else if (!strcmp (usedkey, mainkey))
	_snprintf (desc, size-1,
		   _("You need a passphrase to unlock the secret key for\n"
		     "user: \"%s\"\n"
		     "%s key, ID %s\n"),
		     uid, get_key_pubalgo (algo), usedkey+8);
    safe_free (uid);
    return 0;
}


/* Extract the serial number from the card ID @id and return it. */
const char*
extract_serial_no (const char *id)
{
    static char buf[8];
    char *p;

    p = strchr (id, '/');
    if (!p) {
	log_debug ("extract_serial_no: error '%s'\r\n", id);
	return NULL;
    }
    strncpy (buf, id+(p-id)-6, 6);
    return buf;
}


/* Passphrase callback with the ability to support caching. */
gpgme_error_t
passphrase_cb (void *hook, const char *uid_hint,
	       const char *passphrase_info,
	       int prev_was_bad, int fd)
{
    passphrase_cb_s *c = (passphrase_cb_s*)hook;
    HANDLE hd = (HANDLE)fd;
    void *item;
    const char *keyid=NULL, *pass;
    DWORD n;
    int rc = 0;

    if (!c) {
	log_debug ("passphrase_cb: error '!c'\r\n");
	return gpg_error (GPG_ERR_INV_ARG);
    }
    c->bad_pwd = prev_was_bad? 1 : 0;
    if (prev_was_bad && !c->cancel) {
	if (c->pwd)
	    burn_passphrase (&c->pwd);
	agent_del_cache (c->keyid);
	c->pwd_init = 1;
    }

    if (passphrase_info) {
	if (strlen (passphrase_info) < 16 && 
	    !strstr (passphrase_info, "OPENPGP")) {
	    /* assume symetric encryption. */
	    int pos=2;
	    c->sym.sym_algo = atoi (passphrase_info);
	    if (c->sym.sym_algo > 9)
		pos++;
	    /* XXX: be more strict. */
	    c->sym.s2k_mode = atoi (passphrase_info+pos);
	    c->sym.s2k_hash = atoi (passphrase_info+pos+2);
	}

	keyid = parse_gpg_keyid (passphrase_info);
	pass = agent_get_cache (keyid+8, &item);	
	if (pass) {
	    agent_unlock_cache_entry (&item);
	    c->pwd_init = 0;
	    if (!WriteFile (hd, pass, strlen (pass), &n, NULL))
		log_debug ("passphrase_cb: WriteFile() failed ec=%d\n", w32_errno);
	    if (!WriteFile (hd, "\n", 1, &n, NULL))
		log_debug ("passphrase_cb: WriteFile() failed ec=%d\n", w32_errno);
	    return 0;
	}
    }

    if (c->pwd_init) {
	if (keyid && strlen (keyid) == 16)
	    strcpy (c->keyid, keyid+8);

	/* if @passphrase_info contains 'OPENPGP' we assume a smart card
	    has been used. */
	if (strstr (passphrase_info, "OPENPGP")) {
	    const char *s=passphrase_info;
	    while (s && *s && *s != 'D')
		s++;
	    _snprintf (c->info, sizeof c->info-1,
		       _("Please enter the PIN to unlock your secret card key\n"
		         "Card: %s"), extract_serial_no (s));
	    c->is_card = 1;
	}
	else if (uid_hint)
	    parse_gpg_description (uid_hint, passphrase_info, 
				   c->info, sizeof (c->info) - 1);
	if (c->gpg_cmd == GPG_CMD_DECRYPT) {
	    rc = DialogBoxParam (glob_hinst, (LPCSTR)IDD_WINPT_DECRYPT,
				 (HWND)c->hwnd, passphrase_callback_proc,
				 (LPARAM)c);
	}
	else if (c->gpg_cmd == GPG_CMD_SIGN) {
	    rc = DialogBoxParam (glob_hinst, (LPCSTR)IDD_WINPT_DECRYPT_SIGN,
				 (HWND)c->hwnd, passphrase_callback_proc,
				 (LPARAM)c);
	}
	if (rc == -1) {
	    if (!WriteFile (hd, "\n", 1, &n, NULL))
		log_debug ("passphrase_cb: WriteFile() failed ec=%d\n", w32_errno);
	    return 0;
	}
	c->pwd_init = 0;
    }
    if (c->cancel || !c->pwd) {
	if (!WriteFile (hd, "\n", 1, &n, NULL))
	    log_debug ("passphrase_cb: WriteFile() failed ec=%d\n", w32_errno);
	return 0;
    }

    if (!WriteFile (hd, c->pwd, strlen (c->pwd), &n, NULL))
	log_debug ("passphrase_cb: WriteFile() failed ec=%d\n", w32_errno);
    if (!WriteFile (hd, "\n", 1, &n, NULL))
	log_debug ("passphrase_cb: WriteFile() failed ec=%d\n", w32_errno);
    return 0;
}


/* Initialize the given passphrase callback @cb with the
   used gpgme context @ctx, the command @cmd and a title
   @title for the dialog. */
void
set_gpg_passphrase_cb (passphrase_cb_s *cb, gpgme_ctx_t ctx, 
		       int cmd, HWND hwnd, const char *title)
{
    memset (cb, 0, sizeof *cb);
    cb->gpg_cmd = cmd;
    cb->bad_pwd = 0;
    cb->is_card = 0;
    cb->cancel = 0;
    cb->hwnd = hwnd;
    cb->pwd_init = 1;
    free_if_alloc (cb->title);
    cb->title = m_strdup (title);
    gpgme_set_passphrase_cb (ctx, passphrase_cb, cb);
    cb->gpg = ctx;
}


/* Release the gpg recipient list. */
void
release_gpg_recipients (gpgme_recipient_t *recipients)
{
    gpgme_recipient_t r, n;

    r = *recipients;
    while (r != NULL) {
	n = r->next;
	safe_free (r->keyid);
	safe_free (r);
	r = n;
    }
    *recipients = NULL;
}



/* Release a passphrase callback @ctx. */
void
release_gpg_passphrase_cb (passphrase_cb_s *ctx)
{
    if (!ctx)
	return;
    sfree_if_alloc (ctx->pwd);
    free_if_alloc (ctx->title);
    release_gpg_recipients (&ctx->recipients);
}


/* Simple check to measure passphrase (@pass) quality.
   Return value: 0 on success. */
int
check_passwd_quality (const char *pass, int strict)
{
    int i, nd=0, nc=0, n;

    n = strlen (pass);
    if (n < 8)
	return -1;

    for (i=0; i < n; i++) {
	if (isdigit (pass[i])) 
	    nd++;
	if (isalpha (pass[i])) 
	    nc++;
    }

    /* check that the passphrase contains letters and numbers. */
    if (nd == n || nc == n)
	return -1;

    return 0;
}
