/* wptVerifyList.cpp - Listview for verifying signatures
 *	Copyright (C) 2001, 2002, 2003, 2005 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <time.h>

#include "resource.h"
#include "wptTypes.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptKeylist.h"
#include "wptNLS.h"
#include "wptContext.h"
#include "wptErrors.h"
#include "wptW32API.h"
#include "wptVersion.h"

/* Symbolic column IDs */
enum {
    VER_COL_NAME = 0,
    VER_COL_STAT = 1,
    VER_COL_SIGNED = 2,
    VER_COL_TRUST = 3,
    VER_COL_KEYID = 4,
    VER_COL_UID = 5
};


/* Extract the file name part out of the given path in @path.
   Return value: file part or NULL on error. */
static char*
extract_filename (const char *path)
{
    char * fname, *p;
    int n, len = 0;

    p = strrchr (path, '\\');
    if (!p)
	return m_strdup (path);
    n = p - path;
    len = strlen (path) - n;
    fname = new char[len+1];
    if (!fname)
	BUG (NULL);
    memcpy (fname, path+n+1, len);
    fname[len] = '\0';
    return fname;
}


/* String representaton of the time in @timestamp.
   Format YEAR-MON-DAY HOUR:MIN:SEC.
   Return value: time as formatted string. */
const char*
strtimestamp (long timestamp)
{
    static char timebuf[64];
    struct tm *warp;
    const char *dat;
    
    dat = get_locale_date (timestamp, timebuf, sizeof (timebuf)-1);
    warp = localtime (&timestamp);
    if (!dat)
	_snprintf (timebuf, sizeof (timebuf) - 1, 
		   "%04d-%02d-%02d %02d:%02d:%02d",
		    warp->tm_year+1900, warp->tm_mon+1, warp->tm_mday,	
		    warp->tm_hour, warp->tm_min, warp->tm_sec);
    else
	_snprintf (timebuf, sizeof (timebuf)-1,
		   "%s %02d:%02d:%02d", dat,
		   warp->tm_hour, warp->tm_min, warp->tm_sec);

    return timebuf;
}


/* Return human printable PKA status.
   If no pka information is available, return NULL. */
char*
get_pka_status (gpgme_signature_t sig)
{
    const char *fmt;
    char *pka_inf;

    if (sig->pka_trust == 0 || !sig->pka_address)
	return NULL;
    fmt = _("PKA: Verified signer's address is '%s'");
    pka_inf = new char[strlen (fmt)+strlen (sig->pka_address)+2];
    if (!pka_inf)
	BUG (NULL);	
    sprintf (pka_inf, fmt, sig->pka_address);
    return pka_inf;
}


/* Set additional signature information according to the
   signature @sig. If no info control is used, just return. */
void
verlist_set_additional_info (verlist_ctrl_t vlv, gpgme_signature_t sig)
{
    int used = 1;

    if (!vlv->infctl)
	return;

    if (sig->exp_timestamp > (DWORD)time (NULL))
	SendMessage (vlv->infctl, WM_SETTEXT, 0, (LPARAM)(char*)
		     _("The signature is expired!"));
    else if (sig->validity != GPGME_VALIDITY_MARGINAL &&
	     sig->validity != GPGME_VALIDITY_FULL &&
	     sig->validity != GPGME_VALIDITY_ULTIMATE) {
	SendMessage (vlv->infctl, WM_SETTEXT, 0, (LPARAM)(char*)
		     _("WARNING: This key is not certified with a trusted signature!\r\n"
		       "                   There is no indication that the signature belongs to the owner.\r\n"));
    }
    else {
	char *pka_info = get_pka_status (sig);
	if (pka_info != NULL) {
	    SendMessage (vlv->infctl, WM_SETTEXT, 0, (LPARAM)(char*)pka_info);
	    free_if_alloc (pka_info);
	}
	else {
	    SendMessage (vlv->infctl, WM_SETTEXT, 0, (LPARAM)(char*)"");
	    used = 0;
	}
    }    
    ShowWindow (vlv->infctl, used? SW_SHOW : SW_HIDE);
}


/* Build a verify signature list control. With the parent window
   from @ctrl and the mod given in @fm_mode. @lv contains the
   new control on success.
   Return value: 0 on success. */
void
verlist_build (verlist_ctrl_t *vlv, HWND ctrl, int fm_mode)
{
    struct listview_column_s verlist[] = {
	{0, 120, (char *)_("Name")},
	{1, 140, (char *)_("Status") },
	{2, 120, (char *)_("Signed") },
	{3,  58, (char *)_("Trust") },
	{4,  80, (char *)_("Key ID" )},
	{5, 160, (char *)_("User ID") },
	{6, 0, NULL}
    };
    HICON ico[2];
    struct verlist_ctrl_s *v;
    int j;
    
    v = new verlist_ctrl_s;
    if (!v)
	BUG (0);
    memset (v, 0, sizeof *v);
    listview_new (&v->lv, ctrl);
    for(j=0; verlist[j].fieldname; j++)
	listview_add_column (v->lv, &verlist[j]);
    if (!fm_mode)
	listview_set_column_width (v->lv, 0, 80);
    listview_set_ext_style (v->lv);
    ico[0] = LoadIcon (glob_hinst, (LPCTSTR)IDI_SIG_GOOD);
    ico[1] = LoadIcon (glob_hinst, (LPCTSTR)IDI_SIG_BAD);
    listview_set_image_list (v->lv, 16, 16, ico, 2);
    *vlv = v;
}


void
verlist_set_info_control (verlist_ctrl_t vlv, HWND infctl)
{
    vlv->infctl = infctl;
}


/* Delete the given verify control in @lv. */
void
verlist_delete (verlist_ctrl_t vlv)
{
    if (vlv) {
	listview_release (vlv->lv);
	free_if_alloc (vlv);
    }
}


/* Add the given signature in @sig to the verify control @lv.
   Return value: 0 on success. */
int
verlist_add_sig (verlist_ctrl_t vlv, gpgme_signature_t sig)
{
    listview_ctrl_t lv;
    struct winpt_key_s key;
    u32 key_attr;
    const char *attr;
    char keyid[32+1];
    int is_bad;
    
    is_bad = sig->summary & GPGME_SIGSUM_RED? 1 : 0;
    lv = vlv->lv;
    if (listview_add_item_image (lv, " ", is_bad))
	return WPTERR_GENERAL;
        
    listview_add_sub_item (lv, 0, VER_COL_NAME, "Clipboard");

    memset (&key, 0, sizeof (key));
    winpt_get_pubkey (sig->fpr, &key);
    if (sig->summary == 0 && gpg_err_code (sig->status) == GPG_ERR_NO_ERROR)
	attr = get_gpg_sigstat (GPGME_SIGSUM_GREEN);
    else
	attr = get_gpg_sigstat (sig->summary);
    if (attr)
	listview_add_sub_item (lv, 0, VER_COL_STAT, (char *)attr);
    
    attr = strtimestamp (sig->timestamp);
    listview_add_sub_item (lv, 0, VER_COL_SIGNED, (char *)attr);
    
    attr = _("Unknown");
    if (key.ctx) {
	key_attr = key.ext->uids->validity;
	attr = get_key_trust2 (NULL, key_attr, 0, 0);
    }
    listview_add_sub_item (lv, 0, VER_COL_TRUST, (char *)attr);
    
    attr = get_keyid_from_fpr (sig->fpr);	
    _snprintf (keyid, sizeof keyid -1, "0x%s", attr);
    listview_add_sub_item (lv, 0, VER_COL_KEYID, keyid);
    
    attr = key.ctx? key.ext->uids->name : _("user ID not found");
    listview_add_sub_item (lv, 0, VER_COL_UID, attr);

    if (vlv->infctl)
	verlist_set_additional_info (vlv, sig);
    return 0;
}


/* Add the given file signature in @log to the verify control @lv.
   Return value: 0 on success. */
int
verlist_add_sig_log (verlist_ctrl_t vlv, file_sig_ctx_t log)
{
    gpgme_signature_t sig = log->sig;
    struct listview_ctrl_s *lv;
    struct winpt_key_s key;
    const char *attr;
    char t[64], *name;

    lv = vlv->lv;
    if (listview_add_item (lv, "")) {
	log_debug ("verlist_add_sig_log: listview_add_item() failed.\n");
	return WPTERR_GENERAL;
    }

    memset (&key, 0, sizeof (key));
    winpt_get_pubkey (sig->fpr, &key);
    
    name = extract_filename (log->file);
    if (name)
	listview_add_sub_item (lv, 0, VER_COL_NAME, name);
    else
	listview_add_sub_item (lv, 0, VER_COL_NAME, log->file);
    free_if_alloc (name);
    
    if (sig->summary == 0 && gpg_err_code (sig->status) == GPG_ERR_NO_ERROR)
	attr = get_gpg_sigstat (GPGME_SIGSUM_GREEN);
    else
	attr = get_gpg_sigstat (sig->summary);
    if (attr)
	listview_add_sub_item (lv, 0, VER_COL_STAT, attr);

    if (sig->timestamp > 0)
	attr = strtimestamp (sig->timestamp);
    else
	attr = "No time";
    listview_add_sub_item (lv, 0, VER_COL_SIGNED, attr);
        
    if (key.ctx != NULL)
	attr = get_key_trust2 (NULL, key.ctx->uids->validity, 0, 0);
    else
	attr = _("Unknown");
    listview_add_sub_item (lv, 0, VER_COL_TRUST, attr);

    attr = get_keyid_from_fpr (sig->fpr);
    _snprintf (t, sizeof (t)-1, "0x%s", attr);
    listview_add_sub_item (lv, 0, VER_COL_KEYID, t);
    listview_add_sub_item (lv, 0, VER_COL_UID, 
			   log->user_id? 
			   log->user_id : _("user ID not found"));
    if (vlv->infctl)
	verlist_set_additional_info (vlv, sig);

    return 0;
}
